;; Generated package description from mentor.el  -*- no-byte-compile: t -*-
(define-package "mentor" "0.5" "Frontend for the rTorrent bittorrent client" '((emacs "25.1") (xml-rpc "1.6.15") (seq "1.11") (async "1.9.3") (url-scgi "0.7")) :commit "480602f2ec5467aa2418ee98b328fb85d3b58d0c" :authors '(("Stefan Kangas" . "stefankangas@gmail.com")) :maintainer '("Stefan Kangas" . "stefankangas@gmail.com") :keywords '("comm" "processes" "bittorrent") :url "https://github.com/skangas/mentor")
